#include "portscan.h"
#include <stdarg.h>

int
Sendto (int s, const void *msg, int len, unsigned int flags, const struct
	sockaddr *to, int tolen)
{
	int     resu;

	resu = sendto (s, msg, len, flags, to, tolen);
	if (resu < 0) {
		if (errno == EACCES)	/* indicates broadcast address */
			return (-1);
		perror ("sendto");
		exit (-1);
	}
	return (resu);
}

int
Socket (int domain, int type, int protocol)
{
	int     fd;

	fd = socket (domain, type, protocol);
	if (fd < 0) {
		perror ("socket");
		exit (-1);
	}
	return (fd);
}


int
Print (char *fmt,...)
{
	va_list va;

	va_start (va, fmt);
	if (logfile) {
		vfprintf (logfile, fmt, va);
		fflush (logfile);
	}
	vfprintf (stdout, fmt, va);
	va_end (va);
	return (1);
}

void   *
xmalloc (size_t size)
{
	void   *ptr = malloc (size);

	if (ptr == NULL) {
		Print ("xmalloc: out of memory\n");
		exit (-1);
	}
	return (ptr);
}

char   *
xstrdup (const char *s)
{
	char   *ptr;

	ptr = strdup (s);
	if (ptr == NULL) {
		Print ("xstrdup: out of memory\n");
		exit (-1);
	}
	return (ptr);
}

void   *
xrealloc (void *ptr, size_t size)
{
	void   *tmp;

	tmp = realloc (ptr, size);
	if (tmp == NULL) {
		Print ("xrealloc: out of memory\n");
		exit (-1);
	}
	return (tmp);
}

void
get_linkoffset (void)
{
	int     i;

	i = pcap_datalink (pcap_d);
	switch (i) {
	case DLT_PPP:
		dlt_len = 4;
		break;
	case DLT_RAW:
		dlt_len = 0;
		break;
	case DLT_NULL:
		dlt_len = 4;
		break;
	case DLT_EN10MB:
	default:
		dlt_len = 14;
	}
	return;
}

void
set_filter (char *filter, ...)
{
	va_list		va;
	char		buf[1024];
	bpf_u_int32	netp,
			maskp;
	struct bpf_program fp;

	buf[1023] = '\0';

	va_start (va, filter);
	vsnprintf (buf, sizeof (buf) - 1, filter, va);
	va_end (va);

	if (pcap_lookupnet (opts.interface, &netp, &maskp, errbuf) < 0) {
		fprintf (stderr, "pcap: %s\n", errbuf);
		exit (-1);	
	}

	if (pcap_compile (pcap_d, &fp, buf, 1, maskp) < 0) {
		fprintf (stderr, "pcap: %s\n", errbuf);
		exit (-1);
	}

	if (pcap_setfilter (pcap_d, &fp) < 0) {
		fprintf (stderr, "pcap: %s\n", errbuf);
		exit (-1);
	}
	return;
}
