/* numby - advanced http scanner
 *
 * by scut / team teso
 */

#include <sys/socket.h>
#include <netdb.h>
#include <stdio.h>
#include <time.h>
#include "common.h"
#include "network.h"
#include "numby.h"
#include "http.h"


void
http_skip (int cs)
{
	int	nlcount = 0;
	int	n;
	char	*rline;

	while (nlcount < 1) {
		n = net_rlineta (cs, &rline, 15);
		if (n <= 0)
			return;

		if (strcmp (rline, "\r\n") == 0 || strcmp (rline, "\n") == 0) {
			nlcount += 1;
		} else {
			nlcount = 0;
		}
		free (rline);
	}

	return;
}


int
relay_test (int clientsocket, int serversocket)
{
	int	n = 0;

	if (relay_test_ (clientsocket, serversocket) == 1 &&
		relay_test_ (serversocket, clientsocket))
		n = 1;

	return (n);
}


int
relay_test_ (int clientsocket, int serversocket)
{
	long int	num, rnum;
	char		*response;
	int		n;

	srandom ((unsigned int) time (NULL));

	num = random ();
	net_write (serversocket, "%li\n", num);

	n = net_rlineta (clientsocket, &response, 15);
	if (n <= 0)
		return (0);

	n = sscanf (response, "%li", &rnum);
	free (response);
	if (n != 1 || rnum != num)
		return (0);

	return (1);
}


int
http_gate (setup *set, char *host, unsigned short int port, char *sequence)
{
	int			sd, cs;	/* socket descriptor */
	struct sockaddr_in	sa, csa;
	char			*my_ip, *remote_ip;
	bound			*bnd;

	sd = net_connect (&sa, host, port, NULL, 0, 2);
	if (sd == -1) {
		if (set->machine == 1) {
			printf_r ("0/connection failed/\n");
		} else {
			printf_r ("connect failed\n");
		}
		return (0);
	}

	bnd = net_bind (NULL, 0);
	my_ip = net_getlocalip ();

	net_write (sd, sequence, my_ip, bnd->port);

	cs = net_accept (bnd->bs, &csa, 2);
	if (cs > 0) {
		http_skip (sd);
		if (relay_test (cs, sd) == 1) {
			if (set->machine == 1) {
				printf_r ("1/ok/");
			} else {
				printf_r ("data exchange succeeded /");
			}
		} else {
			if (set->machine == 1) {
				printf_r ("0/connect ok, data exchange failed/");
			} else {
				printf_r ("relay succeeded, data exchange failed /");
			}
		}
		remote_ip = net_peername (cs);
		if (remote_ip != NULL) {
			printf_r ("%s", remote_ip);
			free (remote_ip);
		}
		printf_r ("/\n");
	} else {
		if (set->machine == 1) {
			printf_r ("0/connect ok, relay failed/\n");
		} else {
			printf_r ("relay failed\n");
		}
	}

	net_boundfree (bnd);
	free (my_ip);

	return (1);
}


int
http_connect (setup *set, char *host, unsigned short int port)
{
	return (http_gate (set, host, port, "CONNECT %s:%hu HTTP/1.0\n\n"));
}


int
http_post (setup *set, char *host, unsigned short int port)
{
	return (http_gate (set, host, port, "POST http://%s:%hu/ HTTP/1.0\n\n"));
}


int
http_version (setup *set, char *host, unsigned short int port)
{
	long int		num;
	int			sd, cs;	/* socket descriptor */
	struct sockaddr_in	sa, csa;
	char			*my_ip, *remote_ip;
	bound			*bnd;
	char			*line;

	sd = net_connect (&sa, host, port, NULL, 0, 20);
	if (sd == -1) {
		if (set->machine == 1) {
			printf_r ("0/connection failed/\n");
		} else {
			printf_r ("connect failed\n");
		}
		return (0);
	}

	bnd = net_bind (NULL, 0);
	my_ip = net_getlocalip ();

	srandom ((unsigned int) time (NULL));

	num = random ();
	net_write (sd, "GET http://%s:%hu/%li HTTP/1.0\n\n", my_ip, bnd->port, num);

	cs = net_accept (bnd->bs, &csa, 20);
	if (cs > 0) {
		while (net_rlineta (cs, &line, 10) > 0) {
			if (strncmp (line, "Via:", 4) == 0) {
				if (set->machine == 1) {
					printf_r ("1/%s/", line + 5);
				} else {
					printf_r ("version aquired / %s", line + 5);
				}
			}
			free (line);
		}
		printf_r ("/\n");
	}

	net_boundfree (bnd);
	free (my_ip);

	return (1);

}


int
http_proxy (setup *set, char *host, unsigned short int port)
{
	long int		num;
	int			sd, cs;	/* socket descriptor */
	struct sockaddr_in	sa, csa;
	char			*my_ip, *remote_ip;
	bound			*bnd;

	sd = net_connect (&sa, host, port, NULL, 0, 20);
	if (sd == -1) {
		if (set->machine == 1) {
			printf_r ("0/connection failed/\n");
		} else {
			printf_r ("connect failed\n");
		}
		return (0);
	}

	bnd = net_bind (NULL, 0);
	my_ip = net_getlocalip ();

	srandom ((unsigned int) time (NULL));

	num = random ();
	net_write (sd, "GET http://%s:%hu/%li HTTP/1.0\n\n", my_ip, bnd->port, num);

	cs = net_accept (bnd->bs, &csa, 20);
	if (cs > 0) {
		if (proxy_test (cs, num) == 1) {
			if (set->machine == 1) {
				printf_r ("1/ok/");
			} else {
				printf_r ("proxy request succeeded /");
			}
		} else {
			if (set->machine == 1) {
				printf_r ("0/connect ok, proxy request failed/");
			} else {
				printf_r ("connection succeeded, proxy request failed /");
			}
		}
		remote_ip = net_peername (cs);
		if (remote_ip != NULL) {
			printf_r ("%s", remote_ip);
			free (remote_ip);
		}
		printf_r ("/\n");
	} else {
		if (set->machine == 1) {
			printf_r ("0/connect ok, proxy failed/\n");
		} else {
			printf_r ("proxy failed to establish connection\n");
		}
	}

	net_boundfree (bnd);
	free (my_ip);

	return (1);
}


int
proxy_test (int socket, long int num)
{
	char		*response;
	int		n;
	long int	rnum;

	n = net_rlineta (socket, &response, 20);
	if (n <= 0)
		return (0);

	n = sscanf (response, "GET /%li", &rnum);
	free (response);

	if (n != 1) {
		return (0);
	}

	if (rnum == num)
		return (1);
	return (0);
}



void
http_test (setup *set)
{
	int	pc;	/* portcount */

	for (pc = 0; pc < set->portcount; ++pc) {
		if (set->check_connect == 1) {
			if (set->machine == 1) {
				printf_r ("c/%s:%hu/", set->host, set->port[pc]);
				fflush (stdout);
			} else {
				printf_r ("[ connect] %s:%hu - ", set->host, set->port[pc]);
				fflush (stdout);
			}
			http_connect (set, set->host, set->port[pc]);
		}
		if (set->check_post == 1) {
			if (set->machine == 1) {
				printf_r ("p/%s:%hu/", set->host, set->port[pc]);
				fflush (stdout);
			} else {
				printf_r ("[    post] %s:%hu - ", set->host, set->port[pc]);
				fflush (stdout);
			}
			http_post (set, set->host, set->port[pc]);
		}
		if (set->check_proxy == 1) {
			if (set->machine == 1) {
				printf_r ("x/%s:%hu/", set->host, set->port[pc]);
				fflush (stdout);
			} else {
				printf_r ("[   proxy] %s:%hu - ", set->host, set->port[pc]);
				fflush (stdout);
			}
			http_proxy (set, set->host, set->port[pc]);
		}
		if (set->check_version == 1) {
			if (set->machine == 1) {
				printf_r ("v/%s:%hu/", set->host, set->port[pc]);
				fflush (stdout);
			} else {
				printf_r ("[ version] %s:%hu - ", set->host, set->port[pc]);
				fflush (stdout);
			}
			http_version (set, set->host, set->port[pc]);
		}
	}

	return;
}


