/* numby - advanced http scanner
 *
 * by scut / team teso
 *
 * mainly rewrite of prxtest, yeah :)
 */

#define	NUMBY_MAIN_C

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "network.h"
#include "numby.h"


pthread_mutex_t	out_m;	/* output mutex */


char	*prgname;	/* argv[0] for usage () */


/* local functions
 */

void	usage (void);
void	parse_command (setup *set, char *cmd);
void	parse_ips (setup *set, char *ipstr);
setup	*parse_arg (int argc, char *argv[]);


void
usage (void)
{
	printf ("\n");
	printf ("numby v"VERSION" by "AUTHORS"\n\n");
	printf ("usage: %s [options] [command1 [command2 [...]]]\n"
		"                  <ip>:<port1>[:port2[:port3[...]]]\n\n", prgname);

	printf ("[options]\n");
	printf ("   -m         create machine parseable output\n");
	printf ("   -v         be verbose\n\n");

	printf ("[command]\n");
	printf ("   analyze    do a full analyzation of the http proxy service\n");
	printf ("   azconnect  analyze connect vulnerability\n");
	printf ("   azpost     analyze post vulnerability\n");
	printf ("   azproxy    analyze proxy abilities\n");
	printf ("   version    try to identify http proxy service program\n\n");

	printf ("[required]\n");
	printf ("   ip         the quad dotted ip address or a hostname\n");
	printf ("   port[n]    a unsigned short integer (1-65535) as port number\n\n");

	exit (EXIT_FAILURE);
}


setup *
parse_arg (int argc, char *argv[])
{
	int	argcur;
	setup	*set = xcalloc (1, sizeof (setup));	/* get new setup struct */

	for (argcur = 1; argcur < argc; ++argcur) {

		/* in case of an option
		 */
		if (argv[argcur][0] == '-') {
			switch (argv[argcur][1]) {
			case 'm':
				set->machine = 1;
				break;
			case 'v':
				set->verbose = 1;
				break;
			default:
				usage ();
				break;
			}
		} else if (argcur == (argc - 1)) {
			parse_ips (set, argv[argcur]);
		} else {
			parse_command (set, argv[argcur]);
		}
	}

	return (set);
}


void
parse_command (setup *set, char *cmd)
{
	int		n;
	command		cmds[] = {	{ "analyze", 1, 1, 1, 1 },
					{ "azconnect", 1, 0, 0, 0 },
					{ "azpost", 0, 1, 0, 0 },
					{ "azproxy", 0, 0, 1, 0 },
					{ "version", 0, 0, 0, 1 },
					{ NULL, 0, 0, 0, 0 }
				};


	for (n = 0; cmds[n].commandname != NULL; ++n) {
		if (strcmp (cmd, cmds[n].commandname) == 0) {
			set->check_connect |= cmds[n].flag_connect;
			set->check_post |= cmds[n].flag_post;
			set->check_proxy |= cmds[n].flag_proxy;
			set->check_version |= cmds[n].flag_version;

			return;
		}
	}

	usage ();

	return;
}


void
parse_ips (setup *set, char *ipstr)
{
	char	*sep_ptr;	/* seperator pointer =) */
	char	*cur_ptr;
	int	ip_flag = 0;	/* didn't got ip yet */

	/* strsep should be on any interesting system, therefore we just use
	 * it :)
	 */
	for (sep_ptr = ipstr; (cur_ptr = strsep (&sep_ptr, ":")) != NULL; ++ip_flag) {
		/*break-it-up* *singing*/
		if (ip_flag == 0) {
			set->host = strdup (cur_ptr);
		} else {
			unsigned short int	port;

			set->port = xrealloc (set->port,
				(++set->portcount * sizeof (unsigned short int)));

			if (sscanf (cur_ptr, "%hu", &port) != 1)
				usage ();

			/* save port */
			set->port[(set->portcount - 1)] = port;
		}
	}

	if (set->verbose == 1) {
		int	pc;

		if (set->machine == 1) {
			printf_r ("# new/%s/", set->host);
			for (pc = 0; pc < set->portcount; ++pc) {
				if (pc == 0)
					printf_r ("%hu", set->port[pc]);
				else
					printf_r (",%hu", set->port[pc]);
			}
			printf_r ("/\n");
		} else {
			printf_r ("[    host] %s\n", set->host);
			printf_r ("[   ports]");
			for (pc = 0; pc < set->portcount; ++pc)
				printf_r (" %hu", set->port[pc]);
			printf_r ("\n");
		}
	}

	return;
}


int
main (int argc, char *argv[])
{
	setup	*set;

	prgname = argv[0];
	if (argc < 3)
		usage ();

	pthread_mutex_init (&out_m, NULL);

	set = parse_arg (argc, argv);

	http_test (set);

	exit (EXIT_SUCCESS);
}


