/*
 * telnet do/dont negotiation fingerprinting
 *    by Palmers / teso || pa1mers@gmx.de
 */

#include <telnetfp.hpp>


class main_prog
{
private:
tcp_socket con;
tdfpdb db;
short verbose;

public:

int
switch_verbosity ()
  {
    return (verbose ^= 1);
  }

/*
 * void send_will_wont (unsigned char *):
 * reply to do / dont request with propper will / wont. 
 */
void
send_will_wont (unsigned char *a)
{
  unsigned char will[] = {IAC, WILL, 0, 0},
       wont[] = {IAC, WONT, 0, 0};
  while (strlen ((char *) a) > 0)
    {
      if (a[0] == IAC)
	{
	  if (a[1] == DO)
	    {
	      will[2] = a[2];
	      con.swrite ((char *) will);
	    }
	  else if (a[1] == DONT)
	    {
	      wont[2] = a[2];
	      con.swrite ((char *) wont);
	    }
	}
      a += 3;
    }
}


void
usage (char *s)
{
  printf ("Usage: %s [-v -d <file>] <host>\n", s);
  printf ("\t-v:\t\t turn off verbose output\n");
  printf ("\t-d <file>:\t define from which file finger prints shall be read (default: %s)\n", DEFAULT_DB);
  exit (0);
}


main_prog (int argc, char **argv)
{
  int x = 0;
  printf (PROGRAM VERSION " by "AUTHOR "\n");
  verbose = 1;
  db.init ();
  con.init ();

  if (argc > 1)
    {
      while (argc > x)
        {
	  if (argv[x][0] == '-')
	    {
	       switch (argv[x][1])
		{
		  case 'v':
		    switch_verbosity ();
		    break;
		  case 'd':
		    db.set (argv[x + 1]);
		    break;
		  default:
		    usage (argv[0]);
		}
	    }
	  x++;
        }
    }
  else
    usage (argv[0]);
  check (argv[argc - 1]);
}


int
check (char *host)
{
  unsigned char *do_d = NULL, *dont_d = NULL;
  int x = 0;

  db.open ();
  if (con.sopen (host, 23) != 0)
    {
      printf ("sopen: can not connect\n");
      return 1;
    }
/*
 * 1.: get do's
 */
  do_d = (unsigned char *) con.sread (30);
  if (verbose)
    {
      printf ("DO:   ");
      for (x = 0; x < strlen ((char *) do_d); x++)
        printf ("%d ", do_d[x]);
      printf ("\n");
    }

/*
 * 2.: reply will's
 */
  send_will_wont (do_d);

/*
 * 3.: get dont's
 */
  dont_d = (unsigned char *) con.sread (30);
  if (verbose)
    {
      printf ("DONT: ");
      for (x = 0; x < strlen ((char *) dont_d); x++)
        printf ("%d ", dont_d[x]);
      printf ("\n");
    }

/*
 * 4.: reply wont's
 */
  send_will_wont (dont_d);

  con.sclose ();

/*
 * look up fp, do some output
 */
  db.find_in_db (do_d, dont_d);
  db.close ();  
  return 0;
}
};



int
main (int argc, char **argv)
{
  main_prog a(argc, argv);
}
